#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any, Optional

from jdmn.feel.lib.BaseFEELLib import BaseFEELLib
from jdmn.feel.lib.Types import NUMBER, STRING, BOOLEAN, DATE, TIME, DATE_TIME, DURATION, LIST
from jdmn.feel.lib.type.ComparableComparator import comparable
from jdmn.feel.lib.type.bool.DefaultBooleanLib import DefaultBooleanLib
from jdmn.feel.lib.type.list.DefaultListLib import DefaultListLib
from jdmn.feel.lib.type.numeric.DefaultNumericLib import DefaultNumericLib
from jdmn.feel.lib.type.range.DefaultRangeLib import DefaultRangeLib
from jdmn.feel.lib.type.string.DefaultStringLib import DefaultStringLib
from jdmn.feel.lib.type.time.DefaultDateTimeLib import DefaultDateTimeLib
from jdmn.feel.lib.type.time.DefaultDurationLib import DefaultDurationLib
from jdmn.runtime.DMNRuntimeException import DMNRuntimeException
from jdmn.runtime.LambdaExpression import LambdaExpression
from jdmn.runtime.NumericRoundingMode import NumericRoundingMode
from jdmn.runtime.Range import Range


class BaseStandardFEELLib(BaseFEELLib):
    # Libs
    numberLib = DefaultNumericLib()
    stringLib = DefaultStringLib()
    booleanLib = DefaultBooleanLib()
    dateTimeLib = DefaultDateTimeLib()
    durationLib = DefaultDurationLib()
    listLib = DefaultListLib()
    rangeLib = DefaultRangeLib()

    def __init__(self):
        BaseFEELLib.__init__(self)

    #
    # Conversion functions
    #
    def number(self, literal: STRING, groupingSeparator: STRING = None, decimalSeparator: STRING = None) -> NUMBER:
        try:
            return self.numberLib.number(literal, groupingSeparator, decimalSeparator)
        except Exception as e:
            message = "number({}, {}, {})".format(literal, groupingSeparator, decimalSeparator)
            self.logError(message, e)
            return None

    def string(self, from_: Any) -> STRING:
        try:
            return self.stringLib.string(from_)
        except Exception as e:
            message: STRING = "string({})".format(from_)
            self.logError(message, e)
            return None

    def date(self, *args) -> DATE:
        try:
            return self.dateTimeLib.date(*args)
        except Exception as e:
            message: STRING = "date({})".format(*args)
            self.logError(message, e)
            return None

    # TODO ZoneIDs not supported in ISO 8601
    def time(self, *args) -> TIME:
        try:
            return self.dateTimeLib.time(*args)
        except Exception as e:
            message: STRING = "time({})".format(*args)
            self.logError(message, e)
            return None

    def dateAndTime(self, *args) -> DATE_TIME:
        try:
            return self.dateTimeLib.dateAndTime(*args)
        except Exception as e:
            message: STRING = "dateAndTime({})".format(*args)
            self.logError(message, e)
            return None

    def duration(self, from_: STRING) -> DURATION:
        try:
            return self.durationLib.duration(from_)
        except Exception as e:
            message: STRING = "duration({})".format(from_)
            self.logError(message, e)
            return None

    def yearsAndMonthsDuration(self, from_: DATE, to: DATE) -> DURATION:
        try:
            return self.durationLib.yearsAndMonthsDuration(from_, to)
        except Exception as e:
            message: STRING = "yearsAndMonthsDURATION({}, {})".format(from_, to)
            self.logError(message, e)
            return None

    #
    # Extra conversion functions
    #
    def toDate(self, from_: Any) -> DATE:
        try:
            return self.dateTimeLib.toDate(from_)
        except Exception as e:
            message: STRING = "toDate({})".format(from_)
            self.logError(message, e)
            return None

    def toTime(self, from_: Any) -> TIME:
        try:
            return self.dateTimeLib.toTime(from_)
        except Exception as e:
            message: STRING = "toTime({})".format(from_)
            self.logError(message, e)
            return None

    def toDateTime(self, from_: Any) -> DATE_TIME:
        try:
            return self.dateTimeLib.toDateTime(from_)
        except Exception as e:
            message: STRING = "toTime({})".format(from_)
            self.logError(message, e)
            return None

    #
    # Numeric functions
    #
    def decimal(self, n: NUMBER, scale: NUMBER) -> NUMBER:
        try:
            return self.numberLib.decimal(n, scale)
        except Exception as e:
            message: STRING = "decimal({}, {})".format(n, scale)
            self.logError(message, e)
            return None

    def round(self, n: NUMBER, scale: NUMBER, mode: STRING) -> NUMBER:
        try:
            roundingMode = NumericRoundingMode.fromName(mode)
            if roundingMode is None:
                raise DMNRuntimeException("Unknown rounding mode '{}'. Expected one of '{}'".format(mode, NumericRoundingMode.allowedValues()))
            else:
                return self.numberLib.round(n, scale, roundingMode)

        except Exception as e:
            message: STRING = "round({}, {}, {})".format(n, scale, mode)
            self.logError(message, e)
            return None

    def roundUp(self, n: NUMBER, scale: NUMBER) -> NUMBER:
        try:
            return self.numberLib.round(n, scale, NumericRoundingMode.UP)
        except Exception as e:
            message: STRING = "roundUp({}, {})".format(n, scale)
            self.logError(message, e)
            return None

    def roundDown(self, n: NUMBER, scale: NUMBER) -> NUMBER:
        try:
            return self.numberLib.round(n, scale, NumericRoundingMode.DOWN)
        except Exception as e:
            message: STRING = "roundDown({}, {})".format(n, scale)
            self.logError(message, e)
            return None

    def roundHalfUp(self, n: NUMBER, scale: NUMBER) -> NUMBER:
        try:
            return self.numberLib.round(n, scale, NumericRoundingMode.HALF_UP)
        except Exception as e:
            message: STRING = "roundHalfUp({}, {})".format(n, scale)
            self.logError(message, e)
            return None

    def roundHalfDown(self, n: NUMBER, scale: NUMBER) -> NUMBER:
        try:
            return self.numberLib.round(n, scale, NumericRoundingMode.HALF_DOWN)
        except Exception as e:
            message: STRING = "roundHalfDown({}, {})".format(n, scale)
            self.logError(message, e)
            return None

    def floor(self, *args) -> NUMBER:
        try:
            return self.numberLib.floor(*args)
        except Exception as e:
            message: STRING = "floor({})".format(*args)
            self.logError(message, e)
            return None

    def ceiling(self, *args) -> NUMBER:
        try:
            return self.numberLib.ceiling(*args)
        except Exception as e:
            message: STRING = "ceiling({})".format(*args)
            self.logError(message, e)
            return None

    def abs(self, n: Any) -> Any:
        try:
            if self.isNumber(n):
                return self.numberLib.abs(n)
            else:
                return self.durationLib.abs(n)
        except Exception as e:
            message: STRING = "abs({})".format(n)
            self.logError(message, e)
            return None

    def intModulo(self, dividend: NUMBER, divisor: NUMBER) -> NUMBER:
        try:
            return self.numberLib.intModulo(dividend, divisor)
        except Exception as e:
            message: STRING = "modulo({}, {})".format(dividend, divisor)
            self.logError(message, e)
            return None

    def modulo(self, dividend: NUMBER, divisor: NUMBER) -> NUMBER:
        try:
            return self.numberLib.modulo(dividend, divisor)
        except Exception as e:
            message: STRING = "modulo({}, {})".format(dividend, divisor)
            self.logError(message, e)
            return None

    def sqrt(self, number: NUMBER) -> NUMBER:
        try:
            return self.numberLib.sqrt(number)
        except Exception as e:
            message: STRING = "sqrt({})".format(number)
            self.logError(message, e)
            return None

    def log(self, number: NUMBER) -> NUMBER:
        try:
            return self.numberLib.log(number)
        except Exception as e:
            message: STRING = "log({})".format(number)
            self.logError(message, e)
            return None

    def exp(self, number: NUMBER) -> NUMBER:
        try:
            return self.numberLib.exp(number)
        except Exception as e:
            message: STRING = "exp({})".format(number)
            self.logError(message, e)
            return None

    def odd(self, number: NUMBER) -> BOOLEAN:
        try:
            return self.numberLib.odd(number)
        except Exception as e:
            message: STRING = "odd({})".format(number)
            self.logError(message, e)
            return None

    def even(self, number: NUMBER) -> BOOLEAN:
        try:
            return self.numberLib.even(number)
        except Exception as e:
            message: STRING = "even({})".format(number)
            self.logError(message, e)
            return None

    def mean(self, *args) -> NUMBER:
        try:
            return self.numberLib.mean(*args)
        except Exception as e:
            message: STRING = "mean({})".format(str(args))
            self.logError(message, e)
            return None

    #
    # String functions
    #
    def contains(self, string: STRING, match: STRING) -> BOOLEAN:
        try:
            return self.stringLib.contains(string, match)
        except Exception as e:
            message: STRING = "contains({}, {})".format(string, match)
            self.logError(message, e)
            return None

    def startsWith(self, string: STRING, match: STRING) -> BOOLEAN:
        try:
            return self.stringLib.startsWith(string, match)
        except Exception as e:
            message: STRING = "startsWith({}, {})".format(string, match)
            self.logError(message, e)
            return None

    def endsWith(self, string: STRING, match: STRING) -> BOOLEAN:
        try:
            return self.stringLib.endsWith(string, match)
        except Exception as e:
            message: STRING = "endsWith({}, {})".format(string, match)
            self.logError(message, e)
            return None

    def stringLength(self, string: STRING) -> NUMBER:
        try:
            return None if string is None else self.valueOf(self.stringLib.stringLength(string))
        except Exception as e:
            message: STRING = "stringLength({})".format(string)
            self.logError(message, e)
            return None

    def substring(self, string: STRING, startPosition: NUMBER, length: NUMBER = None) -> STRING:
        try:
            return self.stringLib.substring(string, self.numberLib.toNumber(startPosition), self.numberLib.toNumber(length))
        except Exception as e:
            message: STRING = "substring({}, {}, {})".format(string, startPosition, length)
            self.logError(message, e)
            return None

    def upperCase(self, string: STRING) -> STRING:
        try:
            return self.stringLib.upperCase(string)
        except Exception as e:
            message: STRING = "upperCase({})".format(string)
            self.logError(message, e)
            return None

    def lowerCase(self, string: STRING) -> STRING:
        try:
            return self.stringLib.lowerCase(string)
        except Exception as e:
            message: STRING = "lowerCase({})".format(string)
            self.logError(message, e)
            return None

    def substringBefore(self, string: STRING, match: STRING) -> STRING:
        try:
            return self.stringLib.substringBefore(string, match)
        except Exception as e:
            message: STRING = "substringBefore({}, {})".format(string, match)
            self.logError(message, e)
            return None

    def substringAfter(self, string: STRING, match: STRING) -> STRING:
        try:
            return self.stringLib.substringAfter(string, match)
        except Exception as e:
            message: STRING = "substringAfter({}, {})".format(string, match)
            self.logError(message, e)
            return None

    def replace(self, input: STRING, pattern: STRING, replacement: STRING, flags: STRING = "") -> STRING:
        try:
            return self.stringLib.replace(input, pattern, replacement, flags)
        except Exception as e:
            message: STRING = "replace({}, {}, {}, {})".format(input, pattern, replacement, flags)
            self.logError(message, e)
            return None

    def matches(self, input: STRING, pattern: STRING, flags: STRING = "") -> BOOLEAN:
        try:
            return self.stringLib.matches(input, pattern, flags)
        except Exception as e:
            message: STRING = "matches({}, {}, {})".format(input, pattern, flags)
            self.logError(message, e)
            return None

    def split(self, string: STRING, delimiter: STRING) -> LIST:
        try:
            return self.stringLib.split(string, delimiter)
        except Exception as e:
            message: STRING = "split({}, {})".format(string, delimiter)
            self.logError(message, e)
            return None

    #
    # Boolean functions
    #
    def and_(self, *args) -> BOOLEAN:
        try:
            return self.booleanLib.all(*args)
        except Exception as e:
            message = "and({})".format(*args)
            self.logError(message, e)
            return None

    def all(self, *args) -> BOOLEAN:
        try:
            return self.booleanLib.all(*args)
        except Exception as e:
            message = "and({})".format(*args)
            self.logError(message, e)
            return None

    def or_(self, *args) -> BOOLEAN:
        try:
            return self.booleanLib.any(*args)
        except Exception as e:
            message = "or({})".format(*args)
            self.logError(message, e)
            return None

    def any(self, *args) -> BOOLEAN:
        try:
            return self.booleanLib.any(*args)
        except Exception as e:
            message = "or({})".format(*args)
            self.logError(message, e)
            return None

    def not_(self, operand: BOOLEAN) -> BOOLEAN:
        try:
            return self.booleanType.booleanNot(operand)
        except Exception as e:
            message: STRING = "not({})".format(operand)
            self.logError(message, e)
            return None

    #
    # Date properties
    #
    def year(self, date: DATE) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.year(date))
        except Exception as e:
            message: STRING = "year({})".format(date)
            self.logError(message, e)
            return None

    def month(self, date: DATE) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.month(date))
        except Exception as e:
            message: STRING = "month({})".format(date)
            self.logError(message, e)
            return None

    def day(self, date: DATE) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.day(date))
        except Exception as e:
            message: STRING = "day({})".format(date)
            self.logError(message, e)
            return None

    def weekday(self, date: DATE) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.weekday(date))
        except Exception as e:
            message: STRING = "weekday({})".format(date)
            self.logError(message, e)
            return None

    #
    # Time properties
    #
    def hour(self, time: TIME) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.hour(time))
        except Exception as e:
            message: STRING = "hour({})".format(time)
            self.logError(message, e)
            return None

    def minute(self, time: TIME) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.minute(time))
        except Exception as e:
            message: STRING = "minute({})".format(time)
            self.logError(message, e)
            return None

    def second(self, time: TIME) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.second(time))
        except Exception as e:
            message: STRING = "second({})".format(time)
            self.logError(message, e)
            return None

    def timeOffset(self, time: TIME) -> DURATION:
        try:
            return self.dateTimeLib.timeOffset(time)
        except Exception as e:
            message: STRING = "timeOffset({})".format(time)
            self.logError(message, e)
            return None

    def timezone(self, time: TIME) -> STRING:
        try:
            return self.dateTimeLib.timezone(time)
        except Exception as e:
            message: STRING = "timezone({})".format(time)
            self.logError(message, e)
            return None

    #
    # Duration properties
    #

    def years(self, duration: DURATION) -> NUMBER:
        try:
            return self.valueOf(self.durationLib.years(duration))
        except Exception as e:
            message: STRING = "years({})".format(duration)
            self.logError(message, e)
            return None

    def months(self, duration: DURATION) -> NUMBER:
        try:
            return self.valueOf(self.durationLib.months(duration))
        except Exception as e:
            message: STRING = "months({})".format(duration)
            self.logError(message, e)
            return None

    def days(self, duration: DURATION) -> NUMBER:
        try:
            return self.valueOf(self.durationLib.days(duration))
        except Exception as e:
            message: STRING = "days({})".format(duration)
            self.logError(message, e)
            return None

    def hours(self, duration: DURATION) -> NUMBER:
        try:
            return self.valueOf(self.durationLib.hours(duration))
        except Exception as e:
            message: STRING = "hours({})".format(duration)
            self.logError(message, e)
            return None

    def minutes(self, duration: DURATION) -> NUMBER:
        try:
            return self.valueOf(self.durationLib.minutes(duration))
        except Exception as e:
            message: STRING = "minutes({})".format(duration)
            self.logError(message, e)
            return None

    def seconds(self, duration: DURATION) -> NUMBER:
        try:
            return self.valueOf(self.durationLib.seconds(duration))
        except Exception as e:
            message: STRING = "seconds({})".format(duration)
            self.logError(message, e)
            return None

    #
    # Date and time functions
    #

    def is_(self, value1: Any, value2: Any) -> BOOLEAN:
        try:
            if value1 is None or value2 is None:
                return value1 == value2
            elif type(value1) != type(value2):
                # Different kind
                return False
            elif self.isNumber(value1):
                return self.numericType.numericIs(value1, value2)
            elif self.isBoolean(value1):
                return self.booleanType.booleanIs(value1, value2)
            elif self.isString(value1):
                return self.stringType.stringIs(value1, value2)
            elif self.isDate(value1):
                return self.dateType.dateIs(value1, value2)
            elif self.isTime(value1):
                return self.timeType.timeIs(value1, value2)
            elif self.isDateTime(value1):
                return self.dateTimeType.dateTimeIs(value1, value2)
            elif self.isDuration(value1):
                return self.durationType.durationIs(value1, value2)
            elif self.isList(value1):
                return self.listType.listIs(value1, value2)
            elif self.isRange(value1):
                return self.rangeType.rangeIs(value1, value2)
            elif self.isContext(value1):
                return self.contextType.contextIs(value1, value2)
            else:
                self.logError("'{}' is not supported yet".format(value1.getClass().getSimpleName()))
                return False

        except Exception as e:
            message: STRING = "is({}, {})".format(value1, value2)
            self.logError(message, e)
            return False

    #
    # Temporal functions
    #

    def dayOfYear(self, date: DATE) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.dayOfYear(date))
        except Exception as e:
            message: STRING = "dayOfYear({})".format(date)
            self.logError(message, e)
            return None

    def dayOfWeek(self, date: DATE) -> STRING:
        try:
            return self.dateTimeLib.dayOfWeek(date)
        except Exception as e:
            message: STRING = "dayOfWeek({})".format(date)
            self.logError(message, e)
            return None

    def weekOfYear(self, date: DATE) -> NUMBER:
        try:
            return self.valueOf(self.dateTimeLib.weekOfYear(date))
        except Exception as e:
            message: STRING = "weekOfYear({})".format(date)
            self.logError(message, e)
            return None

    def monthOfYear(self, date: DATE) -> STRING:
        try:
            return self.dateTimeLib.monthOfYear(date)
        except Exception as e:
            message: STRING = "weekOfYear({})".format(date)
            self.logError(message, e)
            return None

    #
    # List functions
    #

    def listContains(self, list: LIST, element: Any) -> BOOLEAN:
        try:
            return self.listLib.listContains(list, element)
        except Exception as e:
            message: STRING = "listContains({}, {})".format(list, element)
            self.logError(message, e)
            return None

    def append(self, list: LIST, *items):
        try:
            return self.listLib.append(list, *items)
        except Exception as e:
            message: STRING = "append({}, {})".format(list, *items)
            self.logError(message, e)
            return None

    def count(self, list: LIST) -> NUMBER:
        try:
            return self.numberLib.count(list)
        except Exception as e:
            message = "count({})".format(list)
            self.logError(message, e)
            return None

    def min(self, *args) -> Optional[comparable]:
        try:
            return self.numberLib.min(*args)
        except Exception as e:
            message = "min({})".format(*args)
            self.logError(message, e)
            return None

    def max(self, *args) -> Optional[comparable]:
        try:
            return self.numberLib.max(*args)
        except Exception as e:
            message = "max({})".format(*args)
            self.logError(message, e)
            return None

    def sum(self, *args) -> Optional[comparable]:
        try:
            return self.numberLib.sum(*args)
        except Exception as e:
            message = "sum({})".format(*args)
            self.logError(message, e)
            return None

    def sublist(self, list: LIST, startPosition: NUMBER, length: NUMBER = None) -> LIST:
        try:
            return self.listLib.sublist(list, self.intValue(startPosition), self.intValue(length))
        except Exception as e:
            message: STRING = "sublist({}, {}, {})".format(list, startPosition, length)
            self.logError(message, e)
            return None

    def concatenate(self, *lists) -> LIST:
        try:
            return self.listLib.concatenate(*lists)
        except Exception as e:
            message: STRING = "concatenate({})".format(str(lists))
            self.logError(message, e)
            return None

    def insertBefore(self, list: LIST, position: NUMBER, newItem: Any) -> LIST:
        try:
            return self.listLib.insertBefore(list, self.intValue(position), newItem)
        except Exception as e:
            message: STRING = "insertBefore({}, {}, {})".format(list, position, newItem)
            self.logError(message, e)
            return None

    def remove(self, list: LIST, position: Any):
        try:
            return self.listLib.remove(list, self.intValue(position))
        except Exception as e:
            message: STRING = "remove({})".format(list)
            self.logError(message, e)
            return None

    def reverse(self, list: LIST) -> LIST:
        try:
            return self.listLib.reverse(list)
        except Exception as e:
            message: STRING = "reverse({})".format(list)
            self.logError(message, e)
            return None

    def indexOf(self, list: LIST, match: Any) -> LIST:
        result = []
        if list is not None:
            for i, o in enumerate(list):
                if o is None and match is None or o is not None and o == match:
                    result.append(self.valueOf(i + 1))
        return result

    def union(self, *lists):
        try:
            return self.listLib.union(*lists)
        except Exception as e:
            message: STRING = "union({})".format(str(lists))
            self.logError(message, e)
            return None

    def distinctValues(self, list: LIST) -> LIST:
        try:
            return self.listLib.distinctValues(list)
        except Exception as e:
            message: STRING = "distinctValues({})".format(list)
            self.logError(message, e)
            return None

    def flatten(self, list: LIST) -> LIST:
        try:
            return self.listLib.flatten(list)
        except Exception as e:
            message: STRING = "flatten({})".format(list)
            self.logError(message, e)
            return None

    def product(self, *args) -> NUMBER:
        try:
            return self.numberLib.product(*args)
        except Exception as e:
            message: STRING = "product({})".format(args)
            self.logError(message, e)
            return None

    def median(self, *args) -> NUMBER:
        try:
            return self.numberLib.median(*args)
        except Exception as e:
            message: STRING = "median({})".format(*args)
            self.logError(message, e)
            return None

    def stddev(self, *args) -> NUMBER:
        try:
            return self.numberLib.stddev(*args)
        except Exception as e:
            message: STRING = "stddev({})".format(*args)
            self.logError(message, e)
            return None

    def mode(self, *args) -> LIST:
        try:
            return self.numberLib.mode(*args)
        except Exception as e:
            message: STRING = "mode({})".format(*args)
            self.logError(message, e)
            return None

    def collect(self, result: LIST, list: LIST) -> None:
        try:
            self.listLib.collect(result, list)
        except Exception as e:
            message: STRING = "collect({}, {})".format(result, list)
            self.logError(message, e)

    def sort(self, list: LIST, precedes: LambdaExpression) -> LIST:
        try:
            return self.listLib.sort(list, precedes)
        except Exception as e:
            message: STRING = "sort({})".format(list)
            self.logError(message, e)
            return None

    #
    # Range functions
    #

    def before(self, arg1: Any, arg2: Any) -> BOOLEAN:
        try:
            return self.rangeLib.before(arg1, arg2)
        except Exception as e:
            message: STRING = "before({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None

    def after(self, arg1: Any, arg2: Any) -> BOOLEAN:
        try:
            return self.rangeLib.after(arg1, arg2)
        except Exception as e:
            message: STRING = "after({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None

    def meets(self, range1: Range, range2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.meets(range1, range2)
        except Exception as e:
            message: STRING = "meets({}, {})".format(range1, range2)
            self.logError(message, e)
            return None

    def metBy(self, range1: Range, range2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.metBy(range1, range2)
        except Exception as e:
            message: STRING = "metBy({}, {})".format(range1, range2)
            self.logError(message, e)
            return None

    def overlaps(self, range1: Range, range2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.overlaps(range1, range2)
        except Exception as e:
            message: STRING = "overlaps({}, {})".format(range1, range2)
            self.logError(message, e)
            return None

    def overlapsBefore(self, range1: Range, range2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.overlapsBefore(range1, range2)
        except Exception as e:
            message: STRING = "overlapsBefore({}, {})".format(range1, range2)
            self.logError(message, e)
            return None

    def overlapsAfter(self, range1: Range, range2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.overlapsAfter(range1, range2)
        except Exception as e:
            message: STRING = "overlapsAfter({}, {})".format(range1, range2)
            self.logError(message, e)
            return None

    def finishes(self, arg1: Any, arg2: Any) -> BOOLEAN:
        try:
            return self.rangeLib.finishes(arg1, arg2)
        except Exception as e:
            message: STRING = "finishes({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None

    def finishedBy(self, arg1: Range, arg2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.finishedBy(arg1, arg2)
        except Exception as e:
            message: STRING = "finishedBy({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None

    def includes(self, arg1: Range, arg2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.includes(arg1, arg2)
        except Exception as e:
            message: STRING = "includes({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None

    def during(self, arg1: Range, arg2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.during(arg1, arg2)
        except Exception as e:
            message: STRING = "during({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None

    def starts(self, arg1: Range, arg2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.starts(arg1, arg2)
        except Exception as e:
            message: STRING = "starts({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None

    def startedBy(self, arg1: Range, arg2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.startedBy(arg1, arg2)
        except Exception as e:
            message: STRING = "startedBy({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None

    def coincides(self, arg1: Range, arg2: Range) -> BOOLEAN:
        try:
            return self.rangeLib.coincides(arg1, arg2)
        except Exception as e:
            message: STRING = "coincides({}, {})".format(arg1, arg2)
            self.logError(message, e)
            return None



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import isodate

from jdmn.feel.lib.Types import STRING, DURATION, DATE, LONG


class DefaultDurationLib:
    def duration(self, from_: STRING) -> DURATION:
        if from_ is None:
            return None

        return isodate.parse_duration(from_)

    def yearsAndMonthsDuration(self, from_: DATE, to: DATE) -> DURATION:
        if from_ is None or to is None:
            return None

        delta = to - from_
        return delta

    def years(self, duration: DURATION) -> LONG:
        if duration is None:
            return None

        return duration.years

    def months(self, duration: DURATION) -> LONG:
        if duration is None:
            return None

        return duration.months

    def days(self, duration: DURATION) -> LONG:
        if duration is None:
            return None

        return duration.days

    def hours(self, duration: DURATION) -> LONG:
        if duration is None:
            return None

        seconds = duration.seconds
        mm, ss = divmod(seconds, 60)
        hh, mm = divmod(mm, 60)
        return hh

    def minutes(self, duration: DURATION) -> LONG:
        if duration is None:
            return None

        seconds = duration.seconds
        mm, ss = divmod(seconds, 60)
        hh, mm = divmod(mm, 60)
        return mm

    def seconds(self, duration: DURATION) -> LONG:
        if duration is None:
            return None

        seconds = duration.seconds
        mm, ss = divmod(seconds, 60)
        return ss

    def abs(self, duration: DURATION) -> DURATION:
        if duration is None:
            return None

        return duration.years



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from functools import cmp_to_key
from typing import Any

from jdmn.feel.lib.Types import BOOLEAN, LIST, INT
from jdmn.runtime.LambdaExpression import LambdaExpression


class DefaultListLib:
    def listContains(self, list_: LIST, element: Any) -> BOOLEAN:
        if list_ is None:
            return None

        return element in list_

    def append(self, list_: LIST, *items) -> LIST:
        result = []
        if list_ is not None:
            result.extend(list_)
        if items is not None:
            for item in items:
                result.append(item)
        else:
            result.append(None)
        return result

    def sublist(self, list_: LIST, position: INT, length: INT = None) -> LIST:
        if position is None:
            return None

        result = []
        if list_ is None or self.isOutOfBounds(list_, position):
            return result
        if position < 0:
            startIndex = len(list_) + position
        else:
            startIndex = position - 1
        endIndex = len(list_) if length is None else startIndex + length
        return list_[startIndex: endIndex]

    def concatenate(self, *lists) -> LIST:
        result = []
        if lists is not None:
            for list_ in lists:
                result.extend(list_)
        return result

    def insertBefore(self, list_: LIST, position: INT, newItem: Any) -> LIST:
        result = []
        if list_ is not None:
            result.extend(list_)
        if self.isOutOfBounds(result, position):
            return result
        if position < 0:
            position = len(result) + position
        else:
            position = position - 1
        result.insert(position, newItem)
        return result

    def remove(self, list_: LIST, position: INT) -> LIST:
        result = []
        if list_ is not None:
            result.extend(list_)
        result.pop(position - 1)
        return result

    def reverse(self, list_: LIST) -> LIST:
        result = []
        if list_ is not None:
            result = list_.copy()
            result.reverse()
        return result

    def union(self, *lists) -> LIST:
        result = []
        if lists is not None:
            for list_ in lists:
                result.extend(list_)
        return self.distinctValues(result)

    def distinctValues(self, list1: LIST) -> LIST:
        result = []
        if list1 is not None:
            for element in list1:
                if element not in result:
                    result.append(element)
        return result

    def flatten(self, list1: LIST) -> LIST:
        if list1 is None:
            return None
        result = []
        self.collect(result, list1)
        return result

    def collect(self, result: LIST, list1: LIST) -> None:
        if list1 is not None:
            for obj in list1:
                if isinstance(obj, list):
                    self.collect(result, obj)
                else:
                    result.append(obj)

    def sort(self, list_: LIST, precedes: LambdaExpression) -> LIST:
        clone = []
        clone.extend(list_)
        clone.sort(key=cmp_to_key(lambda o1, o2: -1 if precedes.apply(o1, o2) else (0 if o1 == o2 else 1)))
        return clone

    @staticmethod
    def isOutOfBounds(list_: LIST, position: INT) -> bool:
        if position is None:
            return False

        length = len(list_)
        if position < 0:
            return not (-length <= position)
        else:
            return not (1 <= position <= length)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import logging
from typing import Any

from jdmn.feel.lib.Types import DATE, TIME, DATE_TIME


class FEELLib:
    LOGGER = logging.getLogger(__name__)

    def logError(self, message: str, e: Exception = None) -> None:
        if e is None:
            self.LOGGER.error(message)
        else:
            self.LOGGER.exception(message)

    def toDate(self, from_: Any) -> DATE:
        raise Exception("Not supported yet")

    def toTime(self, from_: Any) -> TIME:
        raise Exception("Not supported yet")

    def toDateTime(self, from_: Any) -> DATE_TIME:
        raise Exception("Not supported yet")

    #
    # Error conversions
    #
    def toNull(self, obj: Any) -> Any:
        return None



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any

from jdmn.feel.lib.Types import BOOLEAN, RANGE
from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.feel.lib.type.bool.DefaultBooleanType import DefaultBooleanType
from jdmn.runtime.Range import Range


class DefaultRangeType(BaseType):
    def __init__(self):
        BaseType.__init__(self)
        self.booleanType = DefaultBooleanType()

    def isRange(self, value: Any) -> bool:
        return isinstance(value, Range)

    def rangeValue(self, value: RANGE) -> RANGE:
        if isinstance(value, Range):
            return value
        else:
            return None

    def rangeIs(self, r1: RANGE, r2: RANGE) -> BOOLEAN:
        return self.rangeEqual(r1, r2)

    def rangeEqual(self, r1: RANGE, r2: RANGE) -> BOOLEAN:
        if r1 is None and r2 is None:
            return True
        elif r1 is None:
            return False
        elif r2 is None:
            return False
        else:
            return self.isRange(r1) and self.isRange(r2) and r1 == r2

    def rangeNotEqual(self, r1: RANGE, r2: RANGE) -> BOOLEAN:
        return self.booleanType.booleanNot(self.rangeEqual(r1, r2))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Optional

from jdmn.feel.lib.Utils import varArgToList


class DefaultBooleanLib:
    def all(self, *args) -> Optional[bool]:
        operands = varArgToList(*args)

        oneFalse = False
        allTrue = True
        for opd in operands:
            if opd is False:
                oneFalse = True
            if opd is not True:
                allTrue = False
        if oneFalse:
            return False
        elif allTrue:
            return True
        else:
            return None

    def any(self, *args) -> Optional[bool]:
        operands = varArgToList(*args)

        oneTrue = False
        allFalse = True
        for opd in operands:
            if opd is True:
                oneTrue = True
            if opd is not False:
                allFalse = False
        if oneTrue:
            return True
        elif allFalse:
            return False
        else:
            return None



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import decimal
from decimal import Decimal
from typing import Any, Optional, Callable

from jdmn.feel.lib.Types import NUMBER
from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.feel.lib.type.numeric.NumericComparator import NumericComparator


class DefaultNumericType(BaseType):
    DECIMAL128 = decimal.Context(prec=34, rounding=decimal.ROUND_HALF_EVEN)

    def __init__(self):
        BaseType.__init__(self)
        self.numericComparator = NumericComparator()

    def isNumber(self, value: Any) -> bool:
        return isinstance(value, Decimal)

    def numericValue(self, value: NUMBER) -> NUMBER:
        return value

    def numericIs(self, first: NUMBER, second: NUMBER) -> Optional[bool]:
        if first is None or second is None:
            return first == second

        return first == second

    def numericEqual(self, first: NUMBER, second: NUMBER) -> Optional[bool]:
        return self.numericComparator.equalTo(first, second)

    def numericNotEqual(self, first: NUMBER, second: NUMBER) -> Optional[bool]:
        return self.numericComparator.notEqualTo(first, second)

    def numericLessThan(self, first: NUMBER, second: NUMBER) -> Optional[bool]:
        return self.numericComparator.lessThan(first, second)

    def numericGreaterThan(self, first: NUMBER, second: NUMBER) -> Optional[bool]:
        return self.numericComparator.greaterThan(first, second)

    def numericLessEqualThan(self, first: NUMBER, second: NUMBER) -> Optional[bool]:
        return self.numericComparator.lessEqualThan(first, second)

    def numericGreaterEqualThan(self, first: NUMBER, second: NUMBER) -> Optional[bool]:
        return self.numericComparator.greaterEqualThan(first, second)

    def numericAdd(self, first: NUMBER, second: NUMBER) -> Optional[Optional[Decimal]]:
        if first is None or second is None:
            return None

        return self.decimalOperation(first, second, lambda x, y: x + y)

    def numericSubtract(self, first: NUMBER, second: NUMBER) -> Optional[Optional[Decimal]]:
        if first is None or second is None:
            return None

        return self.decimalOperation(first, second, lambda x, y: x - y)

    def numericMultiply(self, first: NUMBER, second: NUMBER) -> Optional[Optional[Decimal]]:
        if first is None or second is None:
            return None

        return self.decimalOperation(first, second, lambda x, y: x * y)

    def numericDivide(self, first: NUMBER, second: NUMBER) -> Optional[Optional[Decimal]]:
        return self.decimalNumericDivide(first, second)

    def numericUnaryMinus(self, first: NUMBER) -> Optional[Optional[Decimal]]:
        if first is None:
            return None

        # DECIMAL 128
        with decimal.localcontext() as ctx:
            ctx.rounding = DefaultNumericType.DECIMAL128.rounding
            ctx.prec = DefaultNumericType.DECIMAL128.prec
            return first.copy_negate()

    def numericExponentiation(self, first: NUMBER, second: NUMBER) -> NUMBER:
        if first is None or second is None:
            return None

        return self.decimalOperation(first, second, lambda x, y: x ** y)

    @staticmethod
    def decimalNumericDivide(first: NUMBER, second: NUMBER) -> NUMBER:
        if first is None or second is None:
            return None
        if second.is_zero():
            return None

        return DefaultNumericType.decimalOperation(first, second, lambda x, y: x / y)

    @staticmethod
    def decimalOperation(first: NUMBER, second: NUMBER, funct: Callable[[NUMBER, NUMBER], NUMBER]):
        # DECIMAL 128
        with decimal.localcontext() as ctx:
            ctx.rounding = DefaultNumericType.DECIMAL128.rounding
            ctx.prec = DefaultNumericType.DECIMAL128.prec
            return funct(first, second)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from datetime import datetime
from typing import Optional

from isodate import Duration

from jdmn.feel.lib.type.time.DefaultCalendarType import DefaultCalendarType
from jdmn.feel.lib.type.time.DefaultDateTimeComparator import DefaultDateTimeComparator


class DefaultDateTimeType(DefaultCalendarType):
    def __init__(self):
        super().__init__()
        self.__dateTimeComparator = DefaultDateTimeComparator()

    def dateTimeIs(self, first: Optional[datetime], second: Optional[datetime]) -> Optional[bool]:
        if first is None or second is None:
            return first == second

        return self.sameDateTime(first, second)

    def dateTimeEqual(self, first: Optional[datetime], second: Optional[datetime]) -> Optional[bool]:
        return self.__dateTimeComparator.equalTo(first, second)

    def dateTimeNotEqual(self, first: Optional[datetime], second: Optional[datetime]) -> Optional[bool]:
        return self.booleanType.booleanNot(self.dateTimeEqual(first, second))

    def dateTimeLessThan(self, first: Optional[datetime], second: Optional[datetime]) -> Optional[bool]:
        return self.__dateTimeComparator.lessThan(first, second)

    def dateTimeGreaterThan(self, first: Optional[datetime], second: Optional[datetime]) -> Optional[bool]:
        return self.__dateTimeComparator.greaterThan(first, second)

    def dateTimeLessEqualThan(self, first: Optional[datetime], second: Optional[datetime]) -> Optional[bool]:
        return self.__dateTimeComparator.lessEqualThan(first, second)

    def dateTimeGreaterEqualThan(self, first: Optional[datetime], second: Optional[datetime]) -> Optional[bool]:
        return self.__dateTimeComparator.greaterEqualThan(first, second)

    def dateTimeSubtract(self, first: Optional[datetime], second: Optional[datetime]) -> Optional[Duration]:
        if first is None or second is None:
            return None

        if self.isDate(first):
            first = self.dateToDateTime(first)
        if self.isDate(second):
            second = self.dateToDateTime(second)
        if self.canNotSubtract(first, second):
            return None

        durationInSeconds = self.dateTimeValue(first) - self.dateTimeValue(second)
        return Duration(seconds=durationInSeconds)

    def dateTimeAddDuration(self, datetime_: Optional[datetime], duration_: Optional[Duration]) -> Optional[datetime]:
        if datetime_ is None or duration_ is None:
            return None

        return datetime_ + duration_

    def dateTimeSubtractDuration(self, datetime_: Optional[datetime], duration_: Optional[Duration]):
        if datetime_ is None or duration_ is None:
            return None

        return self.dateTimeAddDuration(datetime_, - duration_)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any

from jdmn.runtime.listener.Arguments import Arguments
from jdmn.runtime.listener.DRGElement import DRGElement
from jdmn.runtime.listener.EventListener import EventListener
from jdmn.runtime.listener.Rule import Rule


class LoggingEventListener(EventListener):
    def __init__(self, logger):
        self.logger = logger

    def startDRGElement(self, element: DRGElement, arguments: Arguments) -> None:
        self.logger.info("Start {} '{}' with inputs '{}'", element.elementKind.displayName, element.name, arguments)

    def endDRGElement(self, element: DRGElement, arguments: Arguments, output: Any, duration: int) -> None:
        self.logger.info("End {} '{}' with output '{}' in {}ms", element.elementKind.displayName, element.name, output, duration)

    def startRule(self, element: DRGElement, rule: Rule) -> None:
        pass

    def matchRule(self, element: DRGElement, rule: Rule) -> None:
        pass

    def endRule(self, element: DRGElement, rule: Rule, output: Any) -> None:
        self.logger.debug("Rule {} fired with output '{}'", rule.index, output)

    def matchColumn(self, rule: Rule, columnIndex: int, result: Any) -> None:
        self.logger.debug("Test {} checked with output '{}'", columnIndex, result)



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from jdmn.feel.lib.FEELOperatorsTest import FEELOperatorsTest
from jdmn.runtime.Context import Context


class BaseFEELLibTest(FEELOperatorsTest):
    """
    Base test class for BaseFEELLib
    """
    __test__ = False

    #
    # Conversion functions
    #
    def testNumber(self):
        self.assertIsNone(self.getLib().number(None))
        self.assertIsNone(self.getLib().number("1,200.56"))
        self.assertIsNone(self.getLib().number("xxx"))

        self.assertEqualsNumber("123.56", self.getLib().number("123.56"))
        self.assertEqualsNumber("-123.56", self.getLib().number("-123.56"))

    def testDate(self):
        #
        # conversion from string
        #
        self.assertIsNone(self.getLib().date(None))
        self.assertIsNone(self.getLib().date(""))
        self.assertIsNone(self.getLib().date("xxx"))
        self.assertIsNone(self.getLib().date("01211-12-31"))
        self.assertIsNone(self.getLib().date("2017-08-25T11:00:00"))

        # year must be in the range [-999,999,999..999,999,999].
#        self.assertEqualsDateTime("999999999-10-11", self.getLib().date(self.getLib().date("999999999-10-11")))
#        self.assertEqualsDateTime("-999999999-10-11", self.getLib().date(self.getLib().date("-999999999-10-11")))
#        self.assertIsNone(self.getLib().date(self.getLib().date("9999999991-10-11")))
#        self.assertIsNone(self.getLib().date(self.getLib().date("-9999999991-10-11")))

        self.assertEqualsDateTime("2016-08-01", self.getLib().date("2016-08-01"))

        #
        # conversion from date
        #
        self.assertIsNone(self.getLib().date(None))
        self.assertEqualsDateTime("2016-08-01", self.getLib().date(self.makeDate("2016-08-01")))

        #
        # conversion from numbers
        #
        self.assertIsNone(self.getLib().date(None, None, None))
        self.assertIsNone(self.getLib().date(self.makeNumber("2016"), None, None))
        self.assertIsNone(self.getLib().date(None, self.makeNumber("8"), None))
        self.assertIsNone(self.getLib().date(None, None, self.makeNumber("1")))
        self.assertEqualsDateTime("2016-08-01", self.getLib().date(self.makeNumber("2016"), self.makeNumber("8"), self.makeNumber("1")))

    def testTime(self):
        #
        # conversion from string
        #
        self.assertIsNone(self.getLib().time(None))
        self.assertIsNone(self.getLib().time(""))
        self.assertIsNone(self.getLib().time("xxx"))
#        self.assertIsNone(self.getLib().time("13:20:00+01:00@Europe/Paris"))
#        self.assertIsNone(self.getLib().time("13:20:00+00:00[UTC]"))

        # fix input literal
        self.assertEqualsDateTime("11:00:00Z", self.getLib().time("T11:00:00Z"))
        self.assertEqualsDateTime("11:00:00+01:00", self.getLib().time("11:00:00+0100"))

        self.assertEqualsDateTime("11:00:00Z", self.getLib().time("11:00:00Z"))
        self.assertEqualsDateTime("11:00:00.001Z", self.getLib().time("11:00:00.001Z"))

        self.assertEqualsDateTime("11:00:00.001+01:00", self.getLib().time("11:00:00.001+01:00"))
        self.assertEqualsDateTime("11:00:00+01:00", self.getLib().time("11:00:00+01:00"))

        #
        # conversion from number
        #
        self.assertIsNone(self.getLib().time(None, None, None, None))

        self.assertIsNone(self.getLib().time(
            self.makeNumber("12"), self.makeNumber("00"), self.makeNumber("00"),
            self.makeDuration("PT25H10M")))
        self.assertEqualsDateTime("12:00:00+01:10", self.getLib().time(
            self.makeNumber("12"), self.makeNumber("00"), self.makeNumber("00"),
            self.makeDuration("PT1H10M")))

    def testDateAndTime(self):
        #
        # conversion from string
        #
        self.assertIsNone(self.getLib().dateAndTime(None))
        self.assertIsNone(self.getLib().dateAndTime(""))
        self.assertIsNone(self.getLib().dateAndTime("xxx"))
        self.assertIsNone(self.getLib().dateAndTime("11:00:00"))
#        self.assertIsNone(self.getLib().dateAndTime("2011-12-03T10:15:30+01:00@Europe/Paris"))
#        self.assertIsNone(self.getLib().dateAndTime("2017-12-31T12:20:00+19:00"))

        # fix input literal
        self.assertEqualsDateTime("2016-08-01T11:00:00+01:00", self.getLib().dateAndTime("2016-08-01T11:00:00+0100"))

        self.assertEqualsDateTime("2016-08-01T11:00:00Z", self.getLib().dateAndTime("2016-08-01T11:00:00Z"))
        self.assertEqualsDateTime("2016-08-01T11:00:00.001Z", self.getLib().dateAndTime("2016-08-01T11:00:00.001Z"))
        self.assertEqualsDateTime("2016-08-01T11:00:00.001+01:00", self.getLib().dateAndTime("2016-08-01T11:00:00.001+01:00"))
        self.assertEqualsDateTime("2016-08-01T11:00:00+01:00", self.getLib().dateAndTime("2016-08-01T11:00:00+01:00"))

        #
        # conversion from date and time
        #
        self.assertIsNone(self.getLib().dateAndTime(None, None))
        self.assertIsNone(self.getLib().dateAndTime(None, self.makeTime("11:00:00Z")))
        self.assertIsNone(self.getLib().dateAndTime(self.getLib().date("2016-08-01"), None))

        self.assertEqualsDateTime("2016-08-01T11:00:00Z", self.getLib().dateAndTime(self.makeDate("2016-08-01"), self.makeTime("11:00:00Z")))

    def testDuration(self):
        self.assertIsNone(self.getLib().duration("XXX"))
        self.assertIsNone(self.getLib().duration(None))

        self.assertEqualsDateTime("P1Y8M", self.getLib().duration("P1Y8M"))
        self.assertEqualsDateTime("P2DT20H", self.getLib().duration("P2DT20H"))

    #
    # Implicit conversion functions
    #
    def testAsList(self):
        self.assertEqualsList("[None]", self.getLib().asList(None))
        self.assertEqual([], self.getLib().asList())
        self.assertEqual([None, "a"], self.getLib().asList(None, "a"))

    def testAsElement(self):
        self.assertIsNone(self.getLib().asElement(None))
        self.assertIsNone(self.getLib().asElement([]))
        self.assertIsNone(None, self.getLib().asElement(["1", "2"]))

        self.assertEqual("1", self.getLib().asElement(["1"]))

    #
    # Error conversions
    #
    def testToNull(self):
        self.assertIsNone(self.getLib().toNull(None))
        self.assertIsNone(self.getLib().toNull([]))
        self.assertIsNone(self.getLib().toNull(["1", "2"]))

        self.assertIsNone(self.getLib().toNull(["1"]))

    #
    # Extra conversion functions
    #
    def testRangeToList(self):
        self.assertEqual(self.makeNumberList(), self.getLib().rangeToList(False, None, False, None))
        self.assertEqual(self.makeNumberList(), self.getLib().rangeToList(False, None, False, self.makeNumber("3")))
        self.assertEqual(self.makeNumberList(), self.getLib().rangeToList(False, self.makeNumber("1"), False, None))

        self.assertEqual(self.makeNumberList("2"), self.getLib().rangeToList(True, self.makeNumber("1"), True, self.makeNumber("3")))
        self.assertEqual(self.makeNumberList("1", "2"), self.getLib().rangeToList(False, self.makeNumber("1"), True, self.makeNumber("3")))
        self.assertEqual(self.makeNumberList("2", "3"), self.getLib().rangeToList(True, self.makeNumber("1"), False, self.makeNumber("3")))
        self.assertEqual(self.makeNumberList("1", "2", "3"), self.getLib().rangeToList(False, self.makeNumber("1"), False, self.makeNumber("3")))

    def testRangeToListNoFlags(self):
        self.assertEqual([], self.getLib().rangeToList(None, None))
        self.assertEqual([], self.getLib().rangeToList(self.makeNumber("0"), None))
        self.assertEqual([], self.getLib().rangeToList(None, self.makeNumber("1")))

        self.assertEqual(self.makeNumberList(1, 2, 3), self.getLib().rangeToList(self.makeNumber("1"), self.makeNumber("3")))
        self.assertEqual(self.makeNumberList(3, 2, 1), self.getLib().rangeToList(self.makeNumber("3"), self.makeNumber("1")))

    def testToDate(self):
        self.assertIsNone(self.getLib().toDate(None))
        self.assertIsNone(self.getLib().toDate("1"))
        self.assertIsNone(self.getLib().toDate(self.makeNumber("1")))
        self.assertIsNone(self.getLib().toDate(self.makeTime("12:00:00Z")))

        self.assertEqualsDateTime("2016-08-01", self.getLib().toDate(self.makeDate("2016-08-01")))
        self.assertEqualsDateTime("2016-08-01", self.getLib().toDate(self.makeDateAndTime("2016-08-01T12:00:00Z")))

    def testToTime(self):
        self.assertIsNone(self.getLib().toTime(None))
        self.assertIsNone(self.getLib().toTime("1"))
        self.assertIsNone(self.getLib().toTime(self.makeNumber("1")))

        self.assertEqualsDateTime("00:00:00Z", self.getLib().toTime(self.makeDate("2016-08-01")))
        self.assertEqualsDateTime("12:00:00Z", self.getLib().toTime(self.makeTime("12:00:00Z")))
        self.assertEqualsDateTime("12:00:00Z", self.getLib().toTime(self.makeDateAndTime("2016-08-01T12:00:00Z")))

    def testToDateTime(self):
        self.assertIsNone(self.getLib().toDateTime(None))
        self.assertIsNone(self.getLib().toDateTime("1"))
        self.assertIsNone(self.getLib().toDateTime(self.makeNumber("1")))
        self.assertIsNone(self.getLib().toDateTime(self.makeTime("12:00:00Z")))

    #
    # Boolean functions
    #
    def testAnd(self):
        self.assertIsNone(self.getLib().and_(None))
        self.assertTrue(self.getLib().and_([]))
        self.assertIsNone(self.getLib().and_([None, None]))
        self.assertFalse(self.getLib().and_([None, False]))
        self.assertIsNone(self.getLib().and_([None, True]))
        self.assertTrue(self.getLib().and_([True, True]))
        self.assertFalse(self.getLib().and_([True, True, False]))

        self.assertTrue(self.getLib().and_())
        self.assertIsNone(self.getLib().and_(None, None))
        self.assertFalse(self.getLib().and_(None, False))
        self.assertIsNone(self.getLib().and_(None, True))
        self.assertTrue(self.getLib().and_(True, True))
        self.assertFalse(self.getLib().and_(True, True, False))

    def testOr(self):
        self.assertIsNone(self.getLib().or_(None))
        self.assertFalse(self.getLib().or_([]))
        self.assertIsNone(self.getLib().or_([None, None]))
        self.assertIsNone(self.getLib().or_([None, False]))
        self.assertTrue(self.getLib().or_([None, True]))
        self.assertTrue(self.getLib().or_([False, True]))
        self.assertTrue(self.getLib().or_([True, True]))
        self.assertFalse(self.getLib().or_([False, False, False]))

        self.assertFalse(self.getLib().or_())
        self.assertIsNone(self.getLib().or_(None, None))
        self.assertIsNone(self.getLib().or_(None, False))
        self.assertTrue(self.getLib().or_(None, True))
        self.assertTrue(self.getLib().or_(False, True))
        self.assertTrue(self.getLib().or_(True, True))
        self.assertFalse(self.getLib().or_(False, False, False))

    #
    # List functions
    #
    def testElementAt(self):
        self.assertIsNone(self.getLib().elementAt(None, self.makeNumber("1")))
        self.assertIsNone(self.getLib().elementAt(["1", "2", "3"], self.makeNumber("4")))
        self.assertIsNone(self.getLib().elementAt(["1", "2", "3"], self.makeNumber("-4")))
        self.assertIsNone(self.getLib().elementAt(["1", "2", "3"], self.makeNumber("0")))

        self.assertEqual("1", self.getLib().elementAt(["1", "2", "3"], self.makeNumber("1")))
        self.assertEqual("2", self.getLib().elementAt(["1", "2", "3"], self.makeNumber("2")))
        self.assertEqual("3", self.getLib().elementAt(["1", "2", "3"], self.makeNumber("3")))
        self.assertEqual("3", self.getLib().elementAt(["1", "2", "3"], self.makeNumber("-1")))
        self.assertEqual("2", self.getLib().elementAt(["1", "2", "3"], self.makeNumber("-2")))
        self.assertEqual("1", self.getLib().elementAt(["1", "2", "3"], self.makeNumber("-3")))

    def testListContains(self):
        self.assertIsNone(self.getLib().listContains(None, self.makeNumber(2)))
        self.assertIsNone(self.getLib().listContains(None, self.makeNumber(2)))

        self.assertEqual(False, self.getLib().listContains(self.makeNumberList(1, 2, 3), None))
        self.assertEqual(True, self.getLib().listContains(self.makeNumberList(1, 2, 3), self.makeNumber(2)))

    def testFlattenFirstLevel(self):
        self.assertIsNone(self.getLib().flattenFirstLevel(None))

        self.assertEqual([], self.getLib().flattenFirstLevel([]))
        self.assertEqual(["l11", "l12", "l13"], self.getLib().flattenFirstLevel(["l11", "l12", "l13"]))
        self.assertEqual(["l11", "l21", "l22", "l13"], self.getLib().flattenFirstLevel(["l11", ["l21", "l22"], "l13"]))
        self.assertEqual(["l11", "l21", ["l31", "l32"], "l13"], self.getLib().flattenFirstLevel(["l11", ["l21", ["l31", "l32"]], "l13"]))

        self.assertEqual(["l11", None, [None, "l32"], "l13"], self.getLib().flattenFirstLevel(["l11", [None, [None, "l32"]], "l13"]))

    def testCount(self):
        self.assertEqualsNumber(self.makeNumber("0"), self.getLib().count(None))
        self.assertEqualsNumber(self.makeNumber("0"), self.getLib().count(self.makeNumberList()))

        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().count(self.makeNumberList(1, 2, 3)))
        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().count(self.makeNumberList(1, None, 3)))

    def testMin(self):
        self.assertIsNone(self.getLib().min(None))
        self.assertIsNone(self.getLib().min(self.makeNumberList()))
        self.assertIsNone(self.getLib().min(self.makeNumberList(1, None, 3)))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().min(self.makeNumberList(1, 2, 3)))

        self.assertIsNone(self.getLib().min())
        self.assertIsNone(self.getLib().min(self.makeNumber(1), None, self.makeNumber(3)))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().min(self.makeNumber(1), self.makeNumber(2), self.makeNumber(3)))

    def testMax(self):
        self.assertIsNone(self.getLib().max(None))
        self.assertIsNone(self.getLib().max(self.makeNumberList()))
        self.assertIsNone(self.getLib().max(self.makeNumberList(1, None, 3)))
        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().max(self.makeNumberList(1, 2, 3)))

        self.assertIsNone(self.getLib().max())
        self.assertIsNone(self.getLib().max(self.makeNumber(1), None))
        self.assertEqualsNumber(self.makeNumber("3"), self.getLib().max(self.makeNumber(1), self.makeNumber(3)))

    def testSum(self):
        self.assertIsNone(self.getLib().sum(None))
        self.assertIsNone(self.getLib().sum((self.makeNumberList())))
        self.assertIsNone(self.getLib().sum(self.makeNumberList(1, None, 3)))
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().sum(self.makeNumberList(1, 2, 3)))

        self.assertIsNone(self.getLib().sum())
        self.assertIsNone(self.getLib().sum(self.makeNumber(1), None, self.makeNumber(3)))
        self.assertEqualsNumber(self.makeNumber("6"), self.getLib().sum(self.makeNumber(1), self.makeNumber(2), self.makeNumber(3)))

    #
    # Context functions
    #
    def testGetEntries(self):
        self.assertIsNone(self.getLib().getEntries(None))
        self.assertIsNone(self.getLib().getEntries(self.makeNumber("1")))

        self.assertEqual([], self.getLib().getEntries(Context()))
        self.assertEqual([Context().add("key", "a").add("value", self.makeNumber("1"))], self.getLib().getEntries(Context().add("a", self.makeNumber("1"))))

    def testGetValue(self):
        self.assertIsNone(self.getLib().getValue(None, None))
        self.assertIsNone(self.getLib().getValue(Context(), None))
        self.assertIsNone(self.getLib().getValue(Context(), "a"))

        self.assertEqual(self.makeNumber("1"), self.getLib().getValue(Context().add("a", self.makeNumber("1")), "a"))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License")or you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import typing
from datetime import date, time, datetime, timedelta
from decimal import Decimal

from isodate import Duration

from jdmn.runtime.Context import Context
from jdmn.runtime.Range import Range

NUMBER = typing.Optional[Decimal]
STRING = typing.Optional[str]
BOOLEAN = typing.Optional[bool]
DATE = typing.Optional[date]
TIME = typing.Optional[time]
DATE_TIME = typing.Optional[datetime]
DATE_TIME_UNION = typing.Optional[typing.Union[date, time, datetime]]
DURATION = typing.Optional[typing.Union[Duration, timedelta]]
LIST = typing.Optional[typing.List[typing.Any]]
CONTEXT = typing.Optional[Context]
RANGE = typing.Optional[Range]

LONG = typing.Optional[int]
INT = typing.Optional[int]

POINT_RANGE_UNION = typing.Union[typing.Any, RANGE]

COMPARABLE = typing.Optional[typing.Union[Decimal, str, date, time, datetime, Duration, timedelta]]



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any, Optional


class TernaryBooleanLogicUtil:
    def not_(self, operand: Any) -> Optional[bool]:
        if isinstance(operand, bool):
            return not operand
        else:
            return None

    def and_(self, first: Any, second: Any) -> Optional[bool]:
        if self.isBooleanFalse(first) or self.isBooleanFalse(second):
            return False
        elif self.isBooleanTrue(first) and self.isBooleanTrue(second):
            return True
        else:
            return None

    def or_(self, first: Any, second: Any) -> Optional[bool]:
        if self.isBooleanTrue(first) or self.isBooleanTrue(second):
            return True
        elif self.isBooleanFalse(first) and self.isBooleanFalse(second):
            return False
        else:
            return None

    @staticmethod
    def isBooleanTrue(obj: Any) -> bool:
        return isinstance(obj, bool) and obj is True

    @staticmethod
    def isBooleanFalse(obj: Any) -> bool:
        return isinstance(obj, bool) and obj is False



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from datetime import date, datetime, time, timezone, timedelta
from typing import Optional, Any

from isodate import Duration

from jdmn.feel.lib.Types import DATE, TIME, DATE_TIME, INT, DATE_TIME_UNION, DURATION, LONG
from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.feel.lib.type.bool.DefaultBooleanType import DefaultBooleanType


class DefaultCalendarType(BaseType):
    def __init__(self):
        BaseType.__init__(self)
        self.booleanType = DefaultBooleanType()

    @staticmethod
    def isDate(obj: Any) -> bool:
        return isinstance(obj, date) and not isinstance(obj, datetime)

    @staticmethod
    def isTime(obj: Any) -> bool:
        return isinstance(obj, time)

    @staticmethod
    def isDateTime(obj: Any) -> bool:
        return isinstance(obj, datetime)

    def isDuration(self, value: Any):
        return isinstance(value, Duration)

    def isYearsAndMonthsDuration(self, value):
        return isinstance(value, Duration) and value.tdelta.total_seconds() == 0.0

    def isDaysAndTimeDuration(self, value: Any):
        if value is None:
            return False
        else:
            return isinstance(value, timedelta)

    @staticmethod
    def dateToDateTime(date_: DATE) -> datetime:
        return datetime(date_.year, date_.month, date_.day, tzinfo=timezone.utc)

    @staticmethod
    def timeToDateTime(time_: TIME) -> datetime:
        # EPOCH
        return datetime(year=1970, month=1, day=1, hour=time_.hour, minute=time_.minute, second=time_.second, tzinfo=time_.tzinfo)

    @staticmethod
    def sameDate(first: DATE, second: DATE) -> bool:
        return first.year == second.year and first.month == second.month and first.day == second.day

    @staticmethod
    def sameTime(first: TIME, second: TIME) -> bool:
        return first.hour == second.hour and first.minute == second.minute and first.second == second.second and first.tzinfo == second.tzinfo

    def sameDateTime(self, first: DATE_TIME, second: DATE_TIME) -> bool:
        #        return self.sameDate(first.date(), second.date()) and self.sameTime(first.time(), second.time())
        return first.year == second.year and first.month == second.month and \
               first.day == second.day and first.hour == second.hour and first.minute == second.minute and first.second == second.second and first.tzinfo == second.tzinfo

    def value(self, calendar: DATE_TIME_UNION) -> INT:
        if self.isDate(calendar):
            return self.dateValue(calendar)
        elif self.isTime(calendar):
            return self.timeValue(calendar)
        elif self.isDateTime(calendar):
            return self.dateTimeValue(calendar)
        else:
            return None

    def dateValue(self, date_: DATE) -> INT:
        if date_ is None:
            return None

        return self.dateTimeValue(self.dateToDateTime(date_))

    def timeValue(self, time_: TIME) -> INT:
        if time_ is None:
            return None

        value = time_.hour * 3600 + time_.minute * 60 + time_.second
        if time_.tzinfo is not None:
            value -= time_.utcoffset().seconds
        return value

    def dateTimeValue(self, datetime_: DATE_TIME) -> Optional[int]:
        if datetime_ is None:
            return None

        # Seconds from EPOCH
        if datetime_.tzinfo is None:
            # Add UTC
            datetime_ = datetime.combine(datetime_.date(), datetime_.time(), timezone.utc)
        timestamp = datetime_.timestamp()
        return int(timestamp)

    def canNotSubtract(self, first: DATE_TIME, second: DATE_TIME) -> bool:
        return first.tzinfo is None and second.tzinfo is not None or first.tzinfo is not None and second.tzinfo is None

    def durationValue(self, duration: Duration) -> LONG:
        if duration is None:
            return None

        if self.isYearsAndMonthsDuration(duration):
            return self.monthsValue(duration)
        elif self.isDaysAndTimeDuration(duration):
            return self.secondsValue(duration)
        else:
            return int((datetime(1970, 1, 1) + duration).timestamp())

    @staticmethod
    def monthsValue(duration: DURATION) -> LONG:
        if duration is None:
            return None

        years = 0 if duration.years is None else duration.years
        months = 0 if duration.months is None else duration.months
        totalMonths = 12 * years + months
        return totalMonths

    @staticmethod
    def secondsValue(duration: DURATION) -> LONG:
        if duration is None:
            return None

        return int(duration.total_seconds())



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from jdmn.feel.lib.DefaultStandardFEELLib import DefaultStandardFEELLib
from jdmn.feel.lib.Types import COMPARABLE
from jdmn.feel.type.range.AbstractRangeLibTest import AbstractRangeLibTest


class DefaultDateTimeRangeLibTest(AbstractRangeLibTest):
    """
    Base test class for DefaultRangeLib
    """
    __test__ = True

    feelLib = DefaultStandardFEELLib()

    def makePoint(self, number: int) -> COMPARABLE:
        if number < 0 or number > 60:
            raise Exception("Illegal second field")
        return self.feelLib.dateAndTime("2020-01-01T12:00:{0:02d}".format(number))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from typing import Any

from jdmn.feel.lib.Types import BOOLEAN, LIST, CONTEXT
from jdmn.feel.lib.type.BaseType import BaseType
from jdmn.feel.lib.type.bool.DefaultBooleanType import DefaultBooleanType
from jdmn.runtime.Context import Context


class DefaultContextType(BaseType):
    def __init__(self):
        BaseType.__init__(self)
        self.booleanType = DefaultBooleanType()

    def isContext(self, value: Any) -> bool:
        return isinstance(value, Context)

    def contextValue(self, value: CONTEXT) -> CONTEXT:
        if isinstance(value, Context):
            return value
        else:
            return None

    def contextIs(self, c1: CONTEXT, c2: CONTEXT) -> BOOLEAN:
        return self.contextEqual(c1, c2)

    def contextEqual(self, c1: CONTEXT, c2: CONTEXT) -> BOOLEAN:
        if c1 is None and c2 is None:
            return True
        elif c1 is None:
            return False
        elif c2 is None:
            return False
        else:
            return self.isContext(c1) and self.isContext(c2) and c1 == c2

    def contextNotEqual(self, c1: CONTEXT, c2: CONTEXT) -> BOOLEAN:
        return self.booleanType.booleanNot(self.contextEqual(c1, c2))

    def getEntries(self, m: CONTEXT) -> LIST:
        if self.isContext(m):
            result = []
            keys = m.getBindings().keys()
            for key in keys:
                result.append(Context().add("key", key).add("value", m.get(key)))
            return result
        else:
            return None

    def getValue(self, context: CONTEXT, key: Any) -> Any:
        if self.isContext(context):
            return context.get(key)
        else:
            return None



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
import jdmn.runtime.Pair


class PairComparator:
    def compare(self, o1: jdmn.runtime.Pair.Pair, o2: jdmn.runtime.Pair.Pair) -> int:
        priority1 = o1.getRight()
        priority2 = o2.getRight()
        if priority1 is None and priority2 is None:
            return 0
        elif priority1 is None:
            return 1
        elif priority2 is None:
            return -1
        else:
            if isinstance(priority1, int) and isinstance(priority2, int):
                return priority2 - priority1
            else:
                return 1



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use this file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from jdmn.feel.lib.BaseStandardFEELLibTest import BaseStandardFEELLibTest


class DefaultStandardFEELLibTest(BaseStandardFEELLibTest):
    """
    Base test class for DefaultStandardFEELLib
    """
    __test__ = True

    #
    # Time operators
    #
    def testTimeIs(self):
        super().testTimeIs()

        # times with equivalent offset and zone id are not is()
        self.assertFalse(self.getLib().timeIs(self.makeTime("12:00:00"), self.makeTime("12:00:00Z")))
        self.assertFalse(self.getLib().timeIs(self.makeTime("12:00:00"), self.makeTime("12:00:00+00:00")))
        self.assertFalse(self.getLib().timeIs(self.makeTime("00:00:00+00:00"), self.makeTime("00:00:00@Etc/UTC")))
#        self.assertTrue(self.getLib().timeIs(self.makeTime("00:00:00Z"), self.makeTime("00:00:00+00:00")))
        self.assertFalse(self.getLib().timeIs(self.makeTime("00:00:00Z"), self.makeTime("00:00:00@Etc/UTC")))

    #
    # Date time operators
    #
    def testDateTimeIs(self):
        super().testDateTimeIs()

        # datetime with equivalent offset and zone id are not is()
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T12:00:00"), self.makeDateAndTime("2018-12-08T12:00:00+00:00")))
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00+00:00"), self.makeDateAndTime("2018-12-08T00:00:00@Etc/UTC")))
#        self.assertTrue(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T12:00:00Z"), self.makeDateAndTime("2018-12-08T12:00:00+00:00")))
        self.assertFalse(self.getLib().dateTimeIs(self.makeDateAndTime("2018-12-08T00:00:00Z"), self.makeDateAndTime("2018-12-08T00:00:00@Etc/UTC")))

    #
    # Conversion functions
    #
    def testDate(self):
        super().testDate()

        self.assertEqualsDateTime("2016-08-01", self.getLib().date(self.makeDateAndTime("2016-08-01T12:00:00Z")))

    def testTime(self):
        super().testTime()

        #
        # conversion from time, date and date time
        #
        self.assertEqualsDateTime("12:00:00Z", self.getLib().time(self.makeTime("12:00:00Z")))
        self.assertEqualsDateTime("00:00:00Z", self.getLib().time(self.getLib().date("2017-08-10")))
        self.assertEqualsDateTime("11:00:00Z", self.getLib().time(self.makeDateAndTime("2016-08-01T11:00:00Z")))

        #
        # conversion from numbers
        #
        self.assertEqualsDateTime("12:00:00", self.getLib().time(self.makeNumber("12"), self.makeNumber("00"), self.makeNumber("00"), None))

        #
        # conversion from date, time and date time
        #
        self.assertEqualsDateTime("00:00:00Z", self.getLib().time(self.getLib().date("2017-08-10")))
        self.assertEqualsDateTime("12:00:00Z", self.getLib().time(self.makeTime("12:00:00Z")))
        self.assertEqualsDateTime("11:00:00Z", self.getLib().time(self.makeDateAndTime("2016-08-01T11:00:00Z")))

    def testDateAndTime(self):
        super().testDateAndTime()

        #
        # conversion from string
        #
        self.assertEqualsDateTime("2016-08-01T00:00:00", self.getLib().dateAndTime("2016-08-01"))

        # missing Z
        self.assertEqualsDateTime("-2016-01-30T09:05:00", self.getLib().dateAndTime("-2016-01-30T09:05:00"))
        self.assertEqualsDateTime("-2017-02-28T02:02:02", self.getLib().dateAndTime("-2017-02-28T02:02:02"))

        # with zone id
        self.assertEqualsDateTime("2011-12-03T10:15:30@Europe/Paris", self.getLib().dateAndTime("2011-12-03T10:15:30@Europe/Paris"))

        # year must be in the range [-999,999,999..999,999,999]
        self.assertEqualsDateTime("-999999999-12-31T11:22:33", self.getLib().dateAndTime("-999999999-12-31T11:22:33"))
        self.assertEqualsDateTime("999999999-12-31T11:22:33", self.getLib().dateAndTime("999999999-12-31T11:22:33"))
        self.assertIsNone(self.getLib().dateAndTime("-9999999991-12-31T11:22:33"))
        self.assertIsNone(self.getLib().dateAndTime("9999999991-12-31T11:22:33"))

    def testYearsAndMonthsDuration(self):
        super().testYearsAndMonthsDuration()

        self.assertEqualsDateTime("P0Y0M", self.getLib().yearsAndMonthsDuration(self.makeDateAndTime("2015-12-24T12:15:00.000+01:00"),
                                                                                self.makeDateAndTime("2015-12-24T12:15:00.000+01:00")))
        self.assertEqualsDateTime("P1Y2M", self.getLib().yearsAndMonthsDuration(self.makeDateAndTime("2016-09-30T23:25:00"), self.makeDateAndTime("2017-12-28T12:12:12")))
        self.assertEqualsDateTime("P7Y6M", self.getLib().yearsAndMonthsDuration(self.makeDateAndTime("2010-05-30T03:55:58"), self.makeDateAndTime("2017-12-15T00:59:59")))
        self.assertEqualsDateTime("-P4033Y2M", self.getLib().yearsAndMonthsDuration(self.makeDateAndTime("2014-12-31T23:59:59"), self.makeDateAndTime("-2019-10-01T12:32:59")))
        self.assertEqualsDateTime("-P4035Y11M",
                                  self.getLib().yearsAndMonthsDuration(self.makeDateAndTime("2017-09-05T10:20:00-01:00"), self.makeDateAndTime("-2019-10-01T12:32:59+02:00")))

        self.assertEqualsDateTime("P0Y0M", self.getLib().yearsAndMonthsDuration(self.getLib().dateAndTime("2015-12-24T12:15:00.000+01:00"),
                                                                                self.getLib().dateAndTime("2015-12-24T12:15:00.000+01:00")))

    def testString(self):
        self.assertEqual("None", self.getLib().string(None))

        # test number
        self.assertEqual("123.45", self.getLib().string(self.makeNumber("123.45")))

        # test string
        self.assertEqual("True", self.getLib().string(True))

        # test date
        self.assertEqual("2016-08-01", self.getLib().string(self.makeDate("2016-08-01")))
#        self.assertEqual("999999999-12-31", self.getLib().string(self.getLib().date("999999999-12-31")))
#        self.assertEqual("-999999999-12-31", self.getLib().string(self.getLib().date("-999999999-12-31")))
#        self.assertEqual("999999999-12-31", self.getLib().string(self.getLib().date(self.makeNumber(999999999), self.makeNumber(12), self.makeNumber(31))))
#        self.assertEqual("-999999999-12-31", self.getLib().string(self.getLib().date(self.makeNumber(-999999999), self.makeNumber(12), self.makeNumber(31))))

        # test time
        self.assertEqual("11:00:01+00:00", self.getLib().string(self.makeTime("11:00:01Z")))
#        self.assertEqual("11:00:01Z", self.getLib().string(self.makeTime("11:00:01Z")))
#        self.assertEqual("00:01:00@Etc/UTC", self.getLib().string(self.getLib().time("00:01:00@Etc/UTC")))
#        self.assertEqual("00:01:00@Europe/Paris", self.getLib().string(self.getLib().time("00:01:00@Europe/Paris")))
#        self.assertEqual("10:20:00@Europe/Paris", self.getLib().string(self.getLib().time(self.getLib().dateAndTime("2017-08-10T10:20:00@Europe/Paris"))))
#        self.assertEqual("11:20:00@Asia/Dhaka", self.getLib().string(self.getLib().time(self.getLib().dateAndTime("2017-09-04T11:20:00@Asia/Dhaka"))))
#        self.assertEqual("11:59:45+02:45:55",
#                         self.getLib().string(self.getLib().time(self.makeNumber(11), self.makeNumber(59), self.makeNumber(45), self.getLib().duration("PT2H45M55S"))))
#        self.assertEqual("11:59:45-02:45:55",
#                         self.getLib().string(self.getLib().time(self.makeNumber(11), self.makeNumber(59), self.makeNumber(45), self.getLib().duration("-PT2H45M55S"))))
#        self.assertEqual("00:00:00Z", self.getLib().string(self.getLib().time(self.getLib().date("2017-08-10"))))

        # test date time
        self.assertEqual("2016-08-01T11:00:01+00:00", self.getLib().string(self.makeDateAndTime("2016-08-01T11:00:01Z")))
#        self.assertEqual("2016-08-01T11:00:01Z", self.getLib().string(self.makeDateAndTime("2016-08-01T11:00:01Z")))
#        self.assertEqual("99999-12-31T11:22:33", self.getLib().string(self.getLib().dateAndTime("99999-12-31T11:22:33")))
#        self.assertEqual("-99999-12-31T11:22:33", self.getLib().string(self.getLib().dateAndTime("-99999-12-31T11:22:33")))
#        self.assertEqual("2011-12-31T10:15:30@Europe/Paris", self.getLib().string(self.getLib().dateAndTime("2011-12-31T10:15:30@Europe/Paris")))
#        self.assertEqual("2011-12-31T10:15:30@Etc/UTC", self.getLib().string(self.getLib().dateAndTime("2011-12-31T10:15:30@Etc/UTC")))
#        self.assertEqual("2011-12-31T10:15:30.987@Europe/Paris", self.getLib().string(self.getLib().dateAndTime("2011-12-31T10:15:30.987@Europe/Paris")))
#        self.assertEqual("2011-12-31T10:15:30.123456789@Europe/Paris", self.getLib().string(self.getLib().dateAndTime("2011-12-31T10:15:30.123456789@Europe/Paris")))
#        self.assertEqual("999999999-12-31T23:59:59.999999999@Europe/Paris", self.getLib().string(self.getLib().dateAndTime("999999999-12-31T23:59:59.999999999@Europe/Paris")))
#        self.assertEqual("-999999999-12-31T23:59:59.999999999+02:00", self.getLib().string(self.getLib().dateAndTime("-999999999-12-31T23:59:59.999999999+02:00")))
#        self.assertEqual("2017-01-01T23:59:01@Europe/Paris",
#                         self.getLib().string(self.getLib().dateAndTime(self.getLib().date("2017-01-01"), self.getLib().time("23:59:01@Europe/Paris"))))
#        self.assertEqual("2017-01-01T23:59:01.123456789@Europe/Paris",
#                         self.getLib().string(self.getLib().dateAndTime(self.getLib().date("2017-01-01"), self.getLib().time("23:59:01.123456789@Europe/Paris"))))
#        self.assertEqual("2017-09-05T09:15:30.987654321@Europe/Paris",
#                         self.getLib().string(self.getLib().dateAndTime(self.getLib().dateAndTime("2017-09-05T10:20:00"), self.getLib().time("09:15:30.987654321@Europe/Paris"))))
#        self.assertEqual("2017-09-05T09:15:30.987654321@Europe/Paris", self.getLib().string(
#            self.getLib().dateAndTime(self.getLib().dateAndTime("2017-09-05T10:20:00-01:00"), self.getLib().time("09:15:30.987654321@Europe/Paris"))))
#        self.assertEqual("2017-09-05T09:15:30.987654321@Europe/Paris", self.getLib().string(
#            self.getLib().dateAndTime(self.getLib().dateAndTime("2017-09-05T10:20:00@Europe/Paris"), self.getLib().time("09:15:30.987654321@Europe/Paris"))))

    #
    # Time properties
    #
    def testTimeProperties(self):
        self.assertIsNone(self.getLib().hour(None))
        self.assertEqualsNumber(self.makeNumber("12"), self.getLib().hour(self.getLib().time("12:01:02Z")))

        self.assertIsNone(self.getLib().minute(None))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().minute(self.getLib().time("12:01:02Z")))

        self.assertIsNone(self.getLib().second(None))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().second(self.getLib().time("12:01:02Z")))

        self.assertIsNone(self.getLib().timeOffset(None))
        self.assertIsNone(self.getLib().timeOffset(self.getLib().time("12:01:02")))
        self.assertEqual(self.getLib().duration("PT1H"), self.getLib().timeOffset(self.getLib().time("12:01:02+01:00")))
        self.assertEqual(self.getLib().duration("P0Y0M0DT0H0M0.000S"), self.getLib().timeOffset(self.getLib().time("12:01:02Z")))
#        self.assertIsNone(self.getLib().timeOffset(self.getLib().time("12:01:02Z@Etc/UTC")))

        self.assertIsNone(self.getLib().timezone(None))
        self.assertIsNone(self.getLib().timezone(self.getLib().time("12:01:02")))
        self.assertEqual("+01:00", self.getLib().timezone(self.getLib().time("12:01:02+01:00")))
#        self.assertEqual("Z", self.getLib().timezone(self.getLib().time("12:01:02Z")))
#        self.assertEqual("Etc/UTC", self.getLib().timezone(self.getLib().time("12:01:02@Etc/UTC")))
#        self.assertIsNone(self.getLib().timezone(self.getLib().time("12:01:02Z@Etc/UTC")))

    #
    # Date and time properties
    #
    def testDateAndTimeProperties(self):
        self.assertIsNone(self.getLib().year(None))
        self.assertEqualsNumber(self.makeNumber("2018"), self.getLib().year(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))

        self.assertIsNone(self.getLib().month(None))
        self.assertEqualsNumber(self.makeNumber("12"), self.getLib().month(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))

        self.assertIsNone(self.getLib().day(None))
        self.assertEqualsNumber(self.makeNumber("10"), self.getLib().day(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))

        self.assertIsNone(self.getLib().weekday(None))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().weekday(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))

        self.assertIsNone(self.getLib().hour(None))
        self.assertEqualsNumber(self.makeNumber("12"), self.getLib().hour(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))

        self.assertIsNone(self.getLib().minute(None))
        self.assertEqualsNumber(self.makeNumber("1"), self.getLib().minute(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))

        self.assertIsNone(self.getLib().second(None))
        self.assertEqualsNumber(self.makeNumber("2"), self.getLib().second(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))

        self.assertIsNone(self.getLib().timeOffset(None))
        self.assertIsNone(self.getLib().timeOffset(self.getLib().dateAndTime("2018-12-10T12:01:02")))
        self.assertEqual(self.getLib().duration("PT1H"), self.getLib().timeOffset(self.getLib().dateAndTime("2018-12-10T12:01:02+01:00")))
        self.assertEqual(self.getLib().duration("P0Y0M0DT0H0M0.000S"), self.getLib().timeOffset(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))
        self.assertIsNone(self.getLib().timeOffset(self.getLib().dateAndTime("2018-12-10T12:01:02Z@Etc/UTC")))

        self.assertIsNone(self.getLib().timezone(None))
        self.assertIsNone(self.getLib().timezone(self.getLib().dateAndTime("2018-12-10T12:01:02")))
        self.assertEqual("+01:00", self.getLib().timezone(self.getLib().dateAndTime("2018-12-10T12:01:02+01:00")))
#        self.assertEqual("Z", self.getLib().timezone(self.getLib().dateAndTime("2018-12-10T12:01:02Z")))
#        self.assertEqual("Etc/UTC", self.getLib().timezone(self.getLib().dateAndTime("2018-12-10T12:01:02@Etc/UTC")))
        self.assertIsNone(self.getLib().timezone(self.getLib().dateAndTime("2018-12-10T12:01:02Z@Etc/UTC")))



#
# Copyright 2016 Goldman Sachs.
#
# Licensed under the Apache License, Version 2.0 (the "License") you may not use self file except in compliance with the License.
#
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations under the License.
#
from datetime import datetime, time
from typing import Optional

from isodate import Duration

from jdmn.feel.lib.type.time.DefaultCalendarType import DefaultCalendarType
from jdmn.feel.lib.type.time.DefaultTimeComparator import DefaultTimeComparator


class DefaultTimeType(DefaultCalendarType):
    def __init__(self):
        super().__init__()
        self.__timeComparator = DefaultTimeComparator()

    @staticmethod
    def hasTimezone(calendar: datetime) -> bool:
        return calendar.tzinfo is None

    def timeIs(self, first: Optional[time], second: Optional[time]) -> Optional[bool]:
        if first is None or second is None:
            return first == second

        return self.sameTime(first, second)

    def timeEqual(self, first: Optional[time], second: Optional[time]) -> Optional[bool]:
        return self.__timeComparator.equalTo(first, second)

    def timeNotEqual(self, first: Optional[time], second: Optional[time]) -> Optional[bool]:
        return self.booleanType.booleanNot(self.timeEqual(first, second))

    def timeLessThan(self, first: Optional[time], second: Optional[time]) -> Optional[bool]:
        return self.__timeComparator.lessThan(first, second)

    def timeGreaterThan(self, first: Optional[time], second: Optional[time]) -> Optional[bool]:
        return self.__timeComparator.greaterThan(first, second)

    def timeLessEqualThan(self, first: Optional[time], second: Optional[time]) -> Optional[bool]:
        return self.__timeComparator.lessEqualThan(first, second)

    def timeGreaterEqualThan(self, first: Optional[time], second: Optional[time]) -> Optional[bool]:
        return self.__timeComparator.greaterEqualThan(first, second)

    def timeSubtract(self, first: Optional[time], second: Optional[time]) -> Optional[Duration]:
        if first is None or second is None:
            return None

        durationInSeconds = self.timeValue(first) - self.timeValue(second)
        return Duration(seconds=durationInSeconds)

    def timeAddDuration(self, time_: Optional[time], duration_: Optional[Duration]) -> Optional[time]:
        if time_ is None or duration_ is None:
            return None

        return (self.timeToDateTime(time_) + duration_).time()

    def timeSubtractDuration(self, time_: Optional[time], duration_: Optional[Duration]) -> Optional[time]:
        if time_ is None or duration_ is None:
            return None

        return self.timeAddDuration(time_, - duration_)
